#include <stdio.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/termios.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>
#include "sim.h"
#include "table_data.h"

int *regs;

unsigned char *mem;
int rs1, rs2, rd, extra, func;
int maxInstrOp;
int ignoreBpt= 0;  /* Set for first instruction after simulator starts each
		      time (so if we hit a bpt, we don't keep stopping there
		      and never get any further. */
int terminal_width;  /* Used in comms.c for text version of simulator */

int using_x;       /* Using_x is set to 0 if sim is being run on a terminal
		      or something */
int is_tty;

void SigHandler()
{
 gotInt= 1;
 /* *** */
 message("*** Interrupt ***");
}

void ParentCheck()
{
 if (getppid() == 1)  /* Our parent has died */
  exit(0);
}

int init(int msize)
{
 int f;
 static char str[80];
 sprintf(str, "/tmp/DLX.mem.%d", getpid());
 f= open(str, O_RDWR | O_CREAT, 0666);
 if (f < 0)
 {
  /* *** HELP! CAN'T OPEN TMP FILE *** */
  exit(2);
 }
 ftruncate(f, msize + NREGS * sizeof(int));
 mem= (char*)mmap(0, msize + NREGS * sizeof(int), PROT_READ | PROT_WRITE,
		  MAP_SHARED, f, 0);
 if (!mem)
 {
  /* *** HELP! CAN'T MMAP! *** */
  exit(1);
 }
 unlink(str);
 regs= (int*)(mem + msize);  /* Regs are after the memory in the block */
 pc= 0;
 regs[0]= 0;
 maxInstrOp= (sizeof(table) / sizeof(void(*)()))-1;
 /* # of items in jump table */

 gotInt= 0;
 trapNo= -1;
}


void parse(int Type, int Instr)
{
 if (Type == RFMT)
 {
  rs1= (Instr >> 21) & 0x1f;
  rs2= (Instr >> 16) & 0x1f;
  rd=  (Instr >> 11) & 0x1f;
  extra= (Instr >> 6) & 0x1f;
 }
 else if (Type == IFMT)
 {
  rs1= (Instr >> 21) & 0x1f;
  rd=  (Instr >> 16) & 0x1f;
  extra= Instr & 0xffff;
  if (extra & 0x8000)  /* Sign bit, so sign extend into the integer */
   extra |= 0xffff0000;
 }
 else  /* JFMT */
 {
  extra= Instr & 0x3ffffff;
  if (extra & 0x2000000)  /* Sign bit */
   extra |= 0xfc000000;  /* Sign extend */
 }
}

int load(char *name)
{
 extern int sys_nerr;
 extern char *sys_errlist[];
 static char str[128];
 unsigned int addr, value;
 FILE *f= fopen(name, "r");

 if (!f)
 {
  message("Couldn't open %s: %s", name,
	  errno < sys_nerr ? sys_errlist[errno] : "Unknown error");
  return 0;
 }
 while (fgets(str, 128, f))
 {
  /* *** Should read binary *** */
  if (sscanf(str, "%x %x", &addr, &value) < 2)
  {
   message("Not a valid executable file");
   return 0;
  }
  if (addr >= MEM_SIZE)
  {
   message("Attempt to load into nonexistent location %d", addr);
   return 0;
  }
  imem(addr)= value;
 }
 fclose(f);
/* printf("Enter starting pc: ");*/
 pc= 0;
 /* $$$ Maybe something else later... $$$ */
/* { char str[20]; gets(str); sscanf(str, "%x", &pc); oldpc= pc; }*/
 return 1;
}

#ifdef SIM_MAIN
main(int argc, char **argv)
#else
sim_main(int use_x)
#endif
{
 static char str[20];
 int HandleUsr1();
 struct itimerval itv;
 struct sigaction sa;
 sigset_t Empty;
#ifdef SIM_MAIN
 int use_x= 0;
 struct winsize ws;
  ioctl(0, TIOCGWINSZ, &ws);
 terminal_width= ws.ws_col; /* Global var. used by stop() in comms.c */
 
 if (argc > 2)
 {
  message("Usage: %s [program-name]", argv[0]);
  exit(1);  /* *** ? *** */
 }

 init(MEM_SIZE);

 if (argc == 2)
  if (!load(argv[1])) /* Load failed */
  {
   exit(1);  /* *** ? *** */
  }
 is_tty= isatty(0);
#endif
 sigemptyset(&Empty);
 sa.sa_handler= SigHandler;
 sa.sa_mask= Empty;
 sa.sa_flags= SA_RESTART;
 sigaction(ATTN_SIG, &sa, NULL);  /* For asynch comms with X interface */
 sa.sa_handler= HandleUsr1;
 sigaction(SIGUSR1, &sa, NULL);  /* For getting input from kbd (trap 3) */
 sa.sa_handler= ParentCheck;
 sigaction(SIGALRM, &sa, NULL);  /* Make sure our parent is still around */
 itv.it_interval.tv_sec= 1;
 itv.it_interval.tv_usec= 0;
 itv.it_value.tv_sec= 1;
 itv.it_value.tv_usec= 0;
 setitimer(ITIMER_REAL, &itv, NULL);
 
 using_x= use_x;  /* Global variable used by message, stop, handleTrap,
		   anything else? */
 stop();  /* Let interface have a go */

 for (;;)
 {
  int n;
  int op;
  int type;
  int result;

  result=FetchMem(&n, pc);
  if (result == 0 || (result == 2 && ignoreBpt))
   /* Continue if read was successful or we hit a breakpoint but are ignoring
      them. */
  {
   trapNo= -1;  /* In case we did get a bpt when we fetched the instruction */
/*   printf("pc=%x  *pc=%x  r1=%d r2=%d r3=%d r4=%d r5=%d r6=%d\n", pc, imem(pc), regs[1], regs[2], regs[3], regs[4], regs[5], regs[6]);*/
/*   sscanf(str, "%x", &n); */
   op= (n >> 26) & 0x3f;
   /* Work out our internal op code for this instruction */
   type= opTable[op].format;
   if (opTable[op].opCode == SPECIAL)
    op= specialTable[n & 0x3f];
   else
    op= opTable[op].opCode;
/*   printf("%s\n", operationNames[op]);*/
   /* **** imem(pc)= n; **** */
   parse(type, n);
   oldpc= pc;
   pc += 4;
   if (op > maxInstrOp)  /* Invalid */
    do_unimp();  /* oldpc should be restored on trap */
   else
    (table[op])();
  }
  else
   oldpc= pc;
  if (!handleTrap())  /* Check for traps */
  {
   stop();
   ignoreBpt= 1;  /* Don't stop on bpts for next instruction */
  }
  else
   ignoreBpt= 0;  /* Past the first instruction now */
 }
}


/*
Local Variables:
compile-command: "make sim"
End:
*/
